/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
/* $Id: est_genome.c,v 1.1 1997/04/03 13:44:09 badger Exp $
   
   $Log: est_genome.c,v $
 * Revision 1.1  1997/04/03  13:44:09  badger
 * Initial revision
 *
 * Revision 1.6  1997/02/10  14:07:54  rmott
 * fixed bug so that splice sites in REVERSE direction (ie ct/ac rather than gt/ag)
 * are found correctly. Output modified so that splice direction is written
 *
 * Revision 1.4  1997/01/30  17:21:22  rmott
 * fixed bug, and now computes area limit better
 *
 * Revision 1.3  1997/01/30  17:03:45  rmott
 * debugged version with debug statements still in.
 * Fixed problem of not initialising edge properly
 *
 * Revision 1.2  1996/08/07  10:12:05  rmott
 
 * Linear-Space version
 *
   * Revision 1.1  1996/08/01  13:55:42  rmott
   * Initial revision
   *
   
   */

#define MAIN 1

#define BOTH 0
#define FORWARD_ONLY 1
#define REVERSE_ONLY 2

#include<stdio.h>
#include<math.h>
#include"cl.h"
#include"seq_util.h"
#include"lin_align.h"


extern int verbose;
extern int indentation;

void 
make_output( FILE *fp, SEQUENCE *genome, SEQUENCE *est, ge_alignment *ge, int match, int mismatch, int gap_penalty, int intron_penalty, int splice_penalty, int minscore, int align, int width);

int 
main( int argc, char **argv)
{
  SEQUENCE *genome;
  SEQUENCE *splice_sites, *reversed_splice_sites;
  SEQUENCE *est, *reversed_est;
  FILE *gfp;
  FILE *efp;
  char gfile[256];
  char efile[256];
  ge_alignment *fge=NULL, *rge=NULL, *bge=NULL;
  int width=50;
  int match=1;
  int mismatch=1;
  int gap_penalty=2;
  int intron_penalty=40;
  int splice_penalty=20;
  int splice=1;
  int align=0;
  int reverse=0;
  int both=0;
  float megabytes=10.0;
  int minscore=30;
  int shuffles=0, max_score=0;
  int seed;
  int best=1;
  int search_mode;
  char mode[256];

  /* the fasta input files */

  gfile[0] = efile[0] = 0;
  gfp = argfile( "-genome", "r", argc, argv, gfile );
  efp = argfile( "-est", "r", argc, argv, efile );

  /* the alignment penalties */

  getint("-match",&match,argc,argv);
  getint("-mismatch",&mismatch,argc,argv);
  getint("-gap_panalty",&gap_penalty,argc,argv);
  getint("-intron_penalty",&intron_penalty,argc,argv);
  getint("-splice_penalty",&splice_penalty,argc,argv);

  /* the min score for an alignment ot be output */

  getint("-minscore",&minscore,argc,argv);

  /* Look for gt/ag splice sites ? */

  getboolean("-splice",&splice,argc,argv);

  /* Print the alignemt */

  getboolean("-align",&align,argc,argv);
  getint("-width",&width,argc,argv); /* line width of alignment in characters */

  /* mode: This is complicated.

     "forward"   just search forward strands of both sequences
     "reverse"   just search forward of genomic vs reverse of est
     "both"      search forward strand of genomic against forward and reverse 
                 THEN: take the best of these two, and re-align assuming a reversed
		 gene so that the splice sites would be appear as ct/ac. Only output the best
		 alignment unless the flag -nobest is set.

		 Thus THREE alignments are made.

		 The output cordinates are such that the genomic sequence is always in the forward 
		 direction.

		 */
		   
  strcpy(mode,"both");
  getarg("-mode",mode,argc,argv);
  if ( !strcmp(mode,"both")  ) 
    search_mode = BOTH;
  else if (!strcmp(mode,"forward") )
    search_mode = FORWARD_ONLY;
  else if (!strcmp(mode,"reverse") )
    search_mode = REVERSE_ONLY;
  else
    {
      fprintf(stderr,"ERROR: search mode  %s must be one of: both, forward, reverse\n", mode );
      exit(1);
    }

  /* just print the best alignment ? */

  getboolean("-best",&best,argc,argv);

  /* max space in megabytes */

  getfloat("-space",&megabytes,argc,argv);

  /* print debugging info */

  getboolean("-verbose",&verbose,argc,argv);

  /* shuffle the sequences to test for statistical significance this many times */

  getint("-shuffle",&shuffles,argc,argv);
  getseed(&seed,argc,argv); /* random number seed */
  seed = -seed;

  gethelp( argc, argv);

  if ( mismatch < 0 )
    mismatch = -mismatch;
  if ( gap_penalty < 0 )
    gap_penalty = -gap_penalty;
  if ( intron_penalty < 0 )
    intron_penalty = -intron_penalty;
  if ( splice_penalty < 0 )
    splice_penalty = -splice_penalty;

  matinit( match, mismatch, gap_penalty, 0, '-' );

  if ( efp && (gfp && ( genome = get_fasta_sq( gfp ) ) ) )
    {
      /* Make sure we have enough space to hold the genomic sequence */

      if ( megabytes < genome->len*1.5e-6 ) 
	{
	  fprintf(stderr, "WARNING: increasing space from %.3f to %.3f Mb\n", megabytes, 1.5e-6*genome->len );
	  megabytes = 1.5e-6*genome->len;
	}

      /* find the GT/AG splice sites */

      if ( splice )
	splice_sites = find_splice_sites( genome, 1 );
      else
	splice_sites = NULL;

      if ( search_mode == BOTH )
	reversed_splice_sites = find_splice_sites( genome, 0 );
      else
	reversed_splice_sites = NULL;

      /* process each est */
      
      while ( est = get_fasta_sq( efp ) )
	{

	  /* if required, make shuffled comparisons to get statistical significance */

	  if ( shuffles > 0 )
	    {
	      SEQUENCE *shuffled_est = seqdup( est );
	      int n;
	      int score;
	      double mean=0, std=0;
	      ge_alignment *sge;

	      for(n=0;n<shuffles;n++)
		{
		  shuffle_seq( shuffled_est, 1, &seed );
		  sge = non_recursive_est_to_genome( shuffled_est, genome, match, mismatch, gap_penalty, intron_penalty, splice_penalty, splice_sites, 0, 0, DIAGONAL );  
		  score = sge->score;
/*		  fprintf(stderr, "%30.30s\n", shuffled_est->s );
		  fprintf(stderr,"%5d score %d seed %d\n", n, score, seed ); */
		  if ( score > max_score ) 
		    max_score = score;
		  mean += score;
		  std += score*score;
		  free_ge(sge);
		}
	      
	      mean /= shuffles;
	      std = sqrt( (std = shuffles*mean*mean)/(shuffles-1.0) );
	      fprintf(stderr, "shuffles: %d max: %d mean: %.2f std dev: %.2f\n", shuffles, max_score, mean, std );
	      minscore = max_score+1; 
	      free_seq(shuffled_est);
	    }
		  
	  if ( search_mode != REVERSE_ONLY ) /* forward strand */
	      fge = linear_space_est_to_genome( est, genome, match, mismatch, gap_penalty, intron_penalty, splice_penalty, splice_sites, megabytes );
	  else 
	    fge = NULL;

	  if ( search_mode != FORWARD_ONLY ) /* reverse strand */
	    {
	      reversed_est = seqdup(est);
	      complement_seq(reversed_est->s);

	      rge = linear_space_est_to_genome( reversed_est, genome, match, mismatch, gap_penalty, intron_penalty, splice_penalty, splice_sites, megabytes );
	    }
	  else
	    rge = NULL;

	  if ( search_mode == BOTH ) /* search both strands */
	    {
	      if ( fge->score > rge->score ) 
		{ 		  /* redo forward search with reversed splice sites */
		  bge = linear_space_est_to_genome( est, genome, match, mismatch, gap_penalty, intron_penalty, splice_penalty, reversed_splice_sites, megabytes );
		  if (  bge->score > fge->score ) /* probably have a reversed gene */
		    {
		      fprintf(stdout, "Note Best alignment is between forward est and forward genome, but splice  sites imply REVERSED GENE\n");
		      make_output(stdout, genome, est, bge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		      if ( best == 0 ) /* print substandard alignment too */
			make_output(stdout, genome, est, fge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		    }
		  else
		    {
		      printf("Note Best alignment is between forward est and forward genome, and splice  sites imply forward gene\n");
		      make_output(stdout, genome, est, fge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		      if ( best == 0 )
			make_output(stdout, genome, est, bge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		    }
		}
	      else 
		{
		  bge = linear_space_est_to_genome( reversed_est,genome, match, mismatch, gap_penalty, intron_penalty, splice_penalty, reversed_splice_sites, megabytes );
		  if (  bge->score > rge->score ) /* probably have a reversed gene */
		    {
		      fprintf(stdout, "Note Best alignment is between reversed est and forward genome, but splice  sites imply REVERSED GENE\n");
		      make_output(stdout, genome, reversed_est, bge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		      if ( best == 0 ) /* print substandard alignment too */
			make_output(stdout, genome, reversed_est, rge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		    }
		  else
		    {
		      printf("Note Best alignment is between reversed est and forward genome, and splice  sites imply forward gene\n");
		      make_output(stdout, genome, reversed_est, rge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		      if ( best == 0 )
			make_output(stdout, genome, reversed_est, bge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, minscore, align, width);
		    }
		}


	    }
	  if ( bge ) {
	    free_ge(bge);
	    bge = NULL;
	  }
	  if ( rge ) {
	    free_ge(rge);
	    rge = NULL;
	  }
	  if ( fge ) {
	    free_ge(fge);
	    fge = NULL;
	  }

	  free_seq(est);
	  if (reversed_est) 
	    free_seq(reversed_est);
	}

      if ( splice_sites )
	free_seq( splice_sites );

      if ( reversed_splice_sites ) 
	free_seq( reversed_splice_sites );

      free_seq(genome); 
      return 0;
    }
  return 1;
}



void 
make_output( FILE *fp, SEQUENCE *genome, SEQUENCE *est, ge_alignment *ge, int match, int mismatch, int gap_penalty, int intron_penalty, int splice_penalty, int minscore, int align, int width)
{
  if ( ge->score >= minscore )
    {
      blast_style_output( fp, genome, est, ge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, 1, 0 );
      fprintf( fp, "\n");
      blast_style_output( fp, genome, est, ge, match, mismatch, gap_penalty, intron_penalty, splice_penalty, 0, 0 );
	      
      if ( align )
	{
	  fprintf(fp, "\n\n%s vs %s:\n", genome->name, est->name );
	  print_align( fp, genome, est, ge, width );
	}
    }
}
